@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import array
 import gc
 import io
@@ -16,15 +18,12 @@ from typing import (
     Any,
     Iterable,
     Iterator,
-    List,
     NoReturn,
-    Tuple,
-    Type,
     TypeVar,
-    Union,
     cast,
 )
 
+import psutil
 import pytest
 from _pytest.fixtures import SubRequest
 from _pytest.logging import LogCaptureFixture
@@ -87,6 +86,8 @@ if socket.has_ipv6:
     else:
         has_ipv6 = True
 
+skip_ipv6_mark = pytest.mark.skipif(not has_ipv6, reason="IPv6 is not available")
+
 
 @pytest.fixture
 def fake_localhost_dns(monkeypatch: MonkeyPatch) -> None:
@@ -102,11 +103,7 @@ def fake_localhost_dns(monkeypatch: MonkeyPatch) -> None:
 @pytest.fixture(
     params=[
         pytest.param(AddressFamily.AF_INET, id="ipv4"),
-        pytest.param(
-            AddressFamily.AF_INET6,
-            id="ipv6",
-            marks=[pytest.mark.skipif(not has_ipv6, reason="no IPv6 support")],
-        ),
+        pytest.param(AddressFamily.AF_INET6, id="ipv6", marks=[skip_ipv6_mark]),
     ]
 )
 def family(request: SubRequest) -> AnyIPAddressFamily:
@@ -158,13 +155,13 @@ class TestTCPStream:
         sock.close()
 
     @pytest.fixture
-    def server_addr(self, server_sock: socket.socket) -> Tuple[str, int]:
+    def server_addr(self, server_sock: socket.socket) -> tuple[str, int]:
         return server_sock.getsockname()[:2]
 
     async def test_extra_attributes(
         self,
         server_sock: socket.socket,
-        server_addr: Tuple[str, int],
+        server_addr: tuple[str, int],
         family: AnyIPAddressFamily,
     ) -> None:
         async with await connect_tcp(*server_addr) as stream:
@@ -181,7 +178,7 @@ class TestTCPStream:
             assert stream.extra(SocketAttribute.remote_port) == server_addr[1]
 
     async def test_send_receive(
-        self, server_sock: socket.socket, server_addr: Tuple[str, int]
+        self, server_sock: socket.socket, server_addr: tuple[str, int]
     ) -> None:
         async with await connect_tcp(*server_addr) as stream:
             client, _ = server_sock.accept()
@@ -194,7 +191,7 @@ class TestTCPStream:
         assert response == b"halb"
 
     async def test_send_large_buffer(
-        self, server_sock: socket.socket, server_addr: Tuple[str, int]
+        self, server_sock: socket.socket, server_addr: tuple[str, int]
     ) -> None:
         def serve() -> None:
             client, _ = server_sock.accept()
@@ -215,7 +212,7 @@ class TestTCPStream:
         assert response == buffer
 
     async def test_send_eof(
-        self, server_sock: socket.socket, server_addr: Tuple[str, int]
+        self, server_sock: socket.socket, server_addr: tuple[str, int]
     ) -> None:
         def serve() -> None:
             client, _ = server_sock.accept()
@@ -241,7 +238,7 @@ class TestTCPStream:
         assert response == b"\ndlrow ,olleh"
 
     async def test_iterate(
-        self, server_sock: socket.socket, server_addr: Tuple[str, int]
+        self, server_sock: socket.socket, server_addr: tuple[str, int]
     ) -> None:
         def serve() -> None:
             client, _ = server_sock.accept()
@@ -263,13 +260,13 @@ class TestTCPStream:
         assert chunks == [b"bl", b"ah"]
 
     async def test_socket_options(
-        self, family: AnyIPAddressFamily, server_addr: Tuple[str, int]
+        self, family: AnyIPAddressFamily, server_addr: tuple[str, int]
     ) -> None:
         async with await connect_tcp(*server_addr) as stream:
             raw_socket = stream.extra(SocketAttribute.raw_socket)
             assert raw_socket.getsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY) != 0
 
-    @pytest.mark.skipif(not has_ipv6, reason="IPv6 is not available")
+    @skip_ipv6_mark
     @pytest.mark.parametrize(
         "local_addr, expected_client_addr",
         [
@@ -311,12 +308,7 @@ class TestTCPStream:
         "target, exception_class",
         [
             pytest.param(
-                "localhost",
-                ExceptionGroup,
-                id="multi",
-                marks=[
-                    pytest.mark.skipif(not has_ipv6, reason="IPv6 is not available")
-                ],
+                "localhost", ExceptionGroup, id="multi", marks=[skip_ipv6_mark]
             ),
             pytest.param("127.0.0.1", ConnectionRefusedError, id="single"),
         ],
@@ -324,7 +316,7 @@ class TestTCPStream:
     async def test_connection_refused(
         self,
         target: str,
-        exception_class: Union[Type[ExceptionGroup], Type[ConnectionRefusedError]],
+        exception_class: type[ExceptionGroup] | type[ConnectionRefusedError],
         fake_localhost_dns: None,
     ) -> None:
         dummy_socket = socket.socket(AddressFamily.AF_INET6)
@@ -342,7 +334,7 @@ class TestTCPStream:
                 assert isinstance(exception, ConnectionRefusedError)
 
     async def test_receive_timeout(
-        self, server_sock: socket.socket, server_addr: Tuple[str, int]
+        self, server_sock: socket.socket, server_addr: tuple[str, int]
     ) -> None:
         def serve() -> None:
             conn, _ = server_sock.accept()
@@ -359,7 +351,7 @@ class TestTCPStream:
 
                 pytest.fail("The timeout was not respected")
 
-    async def test_concurrent_send(self, server_addr: Tuple[str, int]) -> None:
+    async def test_concurrent_send(self, server_addr: tuple[str, int]) -> None:
         async def send_data() -> NoReturn:
             while True:
                 await stream.send(b"\x00" * 4096)
@@ -374,7 +366,7 @@ class TestTCPStream:
                 exc.match("already writing to")
                 tg.cancel_scope.cancel()
 
-    async def test_concurrent_receive(self, server_addr: Tuple[str, int]) -> None:
+    async def test_concurrent_receive(self, server_addr: tuple[str, int]) -> None:
         async with await connect_tcp(*server_addr) as client:
             async with create_task_group() as tg:
                 tg.start_soon(client.receive)
@@ -387,7 +379,7 @@ class TestTCPStream:
                 finally:
                     tg.cancel_scope.cancel()
 
-    async def test_close_during_receive(self, server_addr: Tuple[str, int]) -> None:
+    async def test_close_during_receive(self, server_addr: tuple[str, int]) -> None:
         async def interrupt() -> None:
             await wait_all_tasks_blocked()
             await stream.aclose()
@@ -398,13 +390,13 @@ class TestTCPStream:
                 with pytest.raises(ClosedResourceError):
                     await stream.receive()
 
-    async def test_receive_after_close(self, server_addr: Tuple[str, int]) -> None:
+    async def test_receive_after_close(self, server_addr: tuple[str, int]) -> None:
         stream = await connect_tcp(*server_addr)
         await stream.aclose()
         with pytest.raises(ClosedResourceError):
             await stream.receive()
 
-    async def test_send_after_close(self, server_addr: Tuple[str, int]) -> None:
+    async def test_send_after_close(self, server_addr: tuple[str, int]) -> None:
         stream = await connect_tcp(*server_addr)
         await stream.aclose()
         with pytest.raises(ClosedResourceError):
@@ -436,7 +428,7 @@ class TestTCPStream:
         server_context: SSLContext,
         client_context: SSLContext,
         server_sock: socket.socket,
-        server_addr: Tuple[str, int],
+        server_addr: tuple[str, int],
     ) -> None:
         def serve() -> None:
             with suppress(socket.timeout):
@@ -464,7 +456,7 @@ class TestTCPStream:
         self,
         server_context: SSLContext,
         server_sock: socket.socket,
-        server_addr: Tuple[str, int],
+        server_addr: tuple[str, int],
     ) -> None:
         thread_exception = None
 
@@ -556,16 +548,8 @@ class TestTCPListener:
         "family",
         [
             pytest.param(AddressFamily.AF_INET, id="ipv4"),
-            pytest.param(
-                AddressFamily.AF_INET6,
-                id="ipv6",
-                marks=[pytest.mark.skipif(not has_ipv6, reason="no IPv6 support")],
-            ),
-            pytest.param(
-                socket.AF_UNSPEC,
-                id="both",
-                marks=[pytest.mark.skipif(not has_ipv6, reason="no IPv6 support")],
-            ),
+            pytest.param(AddressFamily.AF_INET6, id="ipv6", marks=[skip_ipv6_mark]),
+            pytest.param(socket.AF_UNSPEC, id="both", marks=[skip_ipv6_mark]),
         ],
     )
     async def test_accept(self, family: AnyIPAddressFamily) -> None:
@@ -694,6 +678,28 @@ class TestTCPListener:
 
             tg.cancel_scope.cancel()
 
+    @skip_ipv6_mark
+    @pytest.mark.skipif(
+        sys.platform == "win32",
+        reason="Windows does not support interface name suffixes",
+    )
+    async def test_bind_link_local(self) -> None:
+        # Regression test for #554
+        link_local_ipv6_address = next(
+            (
+                addr.address
+                for addresses in psutil.net_if_addrs().values()
+                for addr in addresses
+                if addr.address.startswith("fe80::") and "%" in addr.address
+            ),
+            None,
+        )
+        if link_local_ipv6_address is None:
+            pytest.fail("Could not find a link-local IPv6 interface")
+
+        async with await create_tcp_listener(local_host=link_local_ipv6_address):
+            pass
+
 
 @pytest.mark.skipif(
     sys.platform == "win32", reason="UNIX sockets are not available on Windows"
@@ -704,9 +710,7 @@ class TestUNIXStream:
         return tmp_path_factory.mktemp("unix").joinpath("socket")
 
     @pytest.fixture(params=[False, True], ids=["str", "path"])
-    def socket_path_or_str(
-        self, request: SubRequest, socket_path: Path
-    ) -> Union[Path, str]:
+    def socket_path_or_str(self, request: SubRequest, socket_path: Path) -> Path | str:
         return socket_path if request.param else str(socket_path)
 
     @pytest.fixture
@@ -736,7 +740,7 @@ class TestUNIXStream:
             )
 
     async def test_send_receive(
-        self, server_sock: socket.socket, socket_path_or_str: Union[Path, str]
+        self, server_sock: socket.socket, socket_path_or_str: Path | str
     ) -> None:
         async with await connect_unix(socket_path_or_str) as stream:
             client, _ = server_sock.accept()
@@ -979,9 +983,7 @@ class TestUNIXListener:
         return tmp_path_factory.mktemp("unix").joinpath("socket")
 
     @pytest.fixture(params=[False, True], ids=["str", "path"])
-    def socket_path_or_str(
-        self, request: SubRequest, socket_path: Path
-    ) -> Union[Path, str]:
+    def socket_path_or_str(self, request: SubRequest, socket_path: Path) -> Path | str:
         return socket_path if request.param else str(socket_path)
 
     async def test_extra_attributes(self, socket_path: Path) -> None:
@@ -1004,7 +1006,7 @@ class TestUNIXListener:
                 TypedAttributeLookupError, listener.extra, SocketAttribute.remote_port
             )
 
-    async def test_accept(self, socket_path_or_str: Union[Path, str]) -> None:
+    async def test_accept(self, socket_path_or_str: Path | str) -> None:
         async with await create_unix_listener(socket_path_or_str) as listener:
             client = socket.socket(socket.AF_UNIX)
             client.settimeout(1)
@@ -1077,13 +1079,13 @@ async def test_multi_listener(tmp_path_factory: TempPathFactory) -> None:
         event.set()
         await stream.aclose()
 
-    client_addresses: List[Union[str, IPSockAddrType]] = []
-    listeners: List[Listener] = [await create_tcp_listener(local_host="localhost")]
+    client_addresses: list[str | IPSockAddrType] = []
+    listeners: list[Listener] = [await create_tcp_listener(local_host="localhost")]
     if sys.platform != "win32":
         socket_path = tmp_path_factory.mktemp("unix").joinpath("socket")
         listeners.append(await create_unix_listener(socket_path))
 
-    expected_addresses: List[Union[str, IPSockAddrType]] = []
+    expected_addresses: list[str | IPSockAddrType] = []
     async with MultiListener(listeners) as multi_listener:
         async with create_task_group() as tg:
             tg.start_soon(multi_listener.serve, handle)
